@@ -13,8 +13,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 #ifndef GCONF_GCONF_INTERNALS_H
@@ -29,11 +29,9 @@
 #include <config.h>
 #include <libintl.h>
 
-#define _(String) dgettext (GETTEXT_PACKAGE, String)
-#define N_(String) (String)
-
 #include <glib.h>
 #include <glib/gstdio.h>
+#include <glib/gi18n-lib.h>
 #include "gconf-error.h"
 #include "gconf-value.h"
 #include "gconf-engine.h"
@@ -179,9 +177,6 @@ void   gconf_unquote_string_inplace (gchar        *str,
                                      gchar       **end,
                                      GError      **err);
 
-GConfValue* gconf_value_decode (const gchar *encoded);
-gchar*      gconf_value_encode (GConfValue  *val);
-
 /*
  * List/pair conversion stuff
  */
