@@ -15,8 +15,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 
@@ -49,11 +49,21 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <ctype.h>
-#include <syslog.h>
 #include <time.h>
+#ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
+#endif
 #include <locale.h>
 
+#include <dbus/dbus-glib-lowlevel.h>
+
+#ifdef G_OS_WIN32
+#include <io.h>
+#include <conio.h>
+#define _WIN32_WINNT 0x0500 
+#include <windows.h>
+#endif
+
 /* This makes hash table safer when debugging */
 #ifndef GCONF_ENABLE_DEBUG
 #define safe_g_hash_table_insert g_hash_table_insert
@@ -109,8 +119,6 @@ static GConfDatabase*       obtain_database (GSList *addresses,
 static void                 drop_old_databases (void);
 static gboolean             no_databases_in_use (void);
 
-static void gconf_handle_segv (int signum);
-
 /*
  * Flag indicating that we are shutting down, so return errors
  * on any attempted operation. We do this instead of unregistering with
@@ -338,10 +346,16 @@ gconf_server_load_sources(void)
 
   if (addresses == NULL)
     {      
-      /* Try using the default address xml:readwrite:$(HOME)/.gconf */
-      addresses = g_slist_append(addresses, g_strconcat("xml:readwrite:", g_get_home_dir(), "/.gconf", NULL));
+#ifndef G_OS_WIN32
+      const char *home = g_get_home_dir ();
+#else
+      const char *home = _gconf_win32_get_home_dir ();
+#endif
 
-      gconf_log(GCL_DEBUG, _("No configuration files found, trying to use the default config source `%s'"), (char *)addresses->data);
+      /* Try using the default address xml:readwrite:$(HOME)/.gconf */
+      addresses = g_slist_append(addresses, g_strconcat("xml:readwrite:", home, "/.gconf", NULL));
+
+      gconf_log(GCL_DEBUG, _("No configuration files found. Trying to use the default configuration source `%s'"), (char *)addresses->data);
     }
   
   if (addresses == NULL)
@@ -349,7 +363,7 @@ gconf_server_load_sources(void)
       /* We want to stay alive but do nothing, because otherwise every
          request would result in another failed gconfd being spawned.  
       */
-      gconf_log(GCL_ERR, _("No configuration sources in the source path, configuration won't be saved; edit %s"), GCONF_CONFDIR"/path");
+      gconf_log(GCL_ERR, _("No configuration sources in the source path. Configuration won't be saved; edit %s%s"), GCONF_CONFDIR, "/path");
       /* don't request error since there aren't any addresses */
       sources = gconf_sources_new_from_addresses(NULL, NULL);
 
@@ -362,7 +376,7 @@ gconf_server_load_sources(void)
 
       if (error != NULL)
         {
-          gconf_log(GCL_ERR, _("Error loading some config sources: %s"),
+          gconf_log(GCL_ERR, _("Error loading some configuration sources: %s"),
                     error->message);
 
           g_error_free(error);
@@ -374,7 +388,7 @@ gconf_server_load_sources(void)
       g_assert(sources != NULL);
 
       if (sources->sources == NULL)
-        gconf_log(GCL_ERR, _("No config source addresses successfully resolved, can't load or store config data"));
+        gconf_log(GCL_ERR, _("No configuration source addresses successfully resolved. Can't load or store configuration data"));
     
       tmp = sources->sources;
 
@@ -391,7 +405,7 @@ gconf_server_load_sources(void)
 
       /* In this case, some sources may still return TRUE from their writable() function */
       if (!have_writable)
-        gconf_log(GCL_WARNING, _("No writable config sources successfully resolved, may not be able to save some configuration changes"));
+        gconf_log(GCL_WARNING, _("No writable configuration sources successfully resolved. May be unable to save some configuration changes"));
 
         
       /* Install the sources as the default database */
@@ -399,6 +413,10 @@ gconf_server_load_sources(void)
     }
 }
 
+/*
+ * Signal handlers should not log debug messages as this code is non-reentrant.
+ * Please avoid calling gconf_log in this function.
+ */
 static void
 signal_handler (int signo)
 {
@@ -411,21 +429,10 @@ signal_handler (int signo)
   ++in_fatal;
   
   switch (signo) {
-    /* Fast cleanup only */
-  case SIGSEGV:
-  case SIGBUS:
-  case SIGILL:
-    enter_shutdown ();
-    gconf_log (GCL_ERR,
-               _("Received signal %d, dumping core. Please report a GConf bug."),
-               signo);
-    if (g_getenv ("DISPLAY"))
-      gconf_handle_segv (signo);
-    abort ();
-    break;
-
   case SIGFPE:
+#ifdef SIGPIPE
   case SIGPIPE:
+#endif
     /* Go ahead and try the full cleanup on these,
      * though it could well not work out very well.
      */
@@ -434,11 +441,6 @@ signal_handler (int signo)
     /* let the fatal signals interrupt us */
     --in_fatal;
     
-    gconf_log (GCL_ERR,
-               _("Received signal %d, shutting down abnormally. Please file a GConf bug report."),
-               signo);
-
-
     if (gconf_main_is_running ())
       gconf_main_quit ();
     
@@ -450,28 +452,32 @@ signal_handler (int signo)
     /* let the fatal signals interrupt us */
     --in_fatal;
     
-    gconf_log (GCL_INFO,
-               _("Received signal %d, shutting down cleanly"), signo);
-
     if (gconf_main_is_running ())
       gconf_main_quit ();
     break;
 
+#ifdef SIGHUP
   case SIGHUP:
     --in_fatal;
 
     /* reload sources during next periodic_cleanup() */
     need_db_reload = TRUE;
     break;
+#endif
 
+#ifdef SIGUSR1
   case SIGUSR1:
     --in_fatal;
     
     /* it'd be nice to log a message here but it's not very safe, so */
     gconf_log_debug_messages = !gconf_log_debug_messages;
     break;
+#endif
     
   default:
+#ifndef HAVE_SIGACTION
+    signal (signo, signal_handler);
+#endif
     break;
   }
 }
@@ -482,103 +488,248 @@ gconf_get_poa (void)
   return the_poa;
 }
 
+static const char *
+get_introspection_xml (void)
+{
+  return "<!DOCTYPE node PUBLIC \"-//freedesktop//DTD D-BUS Object Introspection 1.0//EN\"\n"
+         "\"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd\">\n"
+         "<node>\n"
+         "  <interface name=\"org.freedesktop.DBus.Introspectable\">\n"
+         "    <method name=\"Introspect\">\n"
+         "      <arg name=\"introspection_xml\" direction=\"out\" type=\"s\"/>\n"
+         "    </method>\n"
+         "  </interface>\n"
+         "  <interface name=\"org.gnome.GConf\">\n"
+         "    <method name=\"GetIOR\">\n"
+         "      <arg name=\"ior\" direction=\"out\" type=\"s\"/>\n"
+         "    </method>\n"
+         "  </interface>\n"
+         "</node>\n";
+}
+
+static DBusHandlerResult
+bus_message_handler (DBusConnection *connection,
+                     DBusMessage    *message,
+                     void           *user_data)
+{
+  DBusMessage *reply;
+
+  reply = NULL;
+
+  if (dbus_message_is_signal (message,
+                              DBUS_INTERFACE_LOCAL,
+                              "Disconnected"))
+    {
+      gconf_main_quit ();
+      return DBUS_HANDLER_RESULT_HANDLED;
+    }
+  else if (dbus_message_is_method_call (message,
+                                        "org.freedesktop.DBus.Introspectable",
+                                        "Introspect"))
+    {
+      const char *introspection_xml;
+
+      introspection_xml = get_introspection_xml ();
+
+      reply = dbus_message_new_method_return (message);
+      dbus_message_append_args (reply, DBUS_TYPE_STRING, &introspection_xml,
+                                DBUS_TYPE_INVALID);
+
+    }
+  else if (dbus_message_is_method_call (message,
+                                        "org.gnome.GConf",
+                                        "GetIOR"))
+    {
+      const char *ior;
+
+      ior = gconf_get_daemon_ior ();
+
+      reply = dbus_message_new_method_return (message);
+      dbus_message_append_args (reply, DBUS_TYPE_STRING, &ior, DBUS_TYPE_INVALID);
+    }
+
+  if (reply != NULL)
+    {
+      dbus_connection_send (connection, reply, NULL);
+      dbus_message_unref (reply);
+      return DBUS_HANDLER_RESULT_HANDLED;
+    }
+
+  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
+static DBusConnection *
+get_on_d_bus (void)
+{
+  DBusConnection *connection;
+  DBusError bus_error;
+  int result;
+
+  dbus_error_init (&bus_error);
+  connection = dbus_bus_get (DBUS_BUS_SESSION, &bus_error);
+
+  if (dbus_error_is_set (&bus_error))
+    {
+      gconf_log (GCL_ERR, _("Could not connect to session bus: %s"), bus_error.message);
+      dbus_error_free (&bus_error);
+      return NULL;
+    }
+
+  dbus_connection_setup_with_g_main (connection, NULL);
+
+  if (!dbus_connection_add_filter (connection, (DBusHandleMessageFunction)
+                                  bus_message_handler, NULL, NULL))
+    {
+      dbus_connection_unref (connection);
+      return NULL;
+    }
+
+  dbus_connection_set_exit_on_disconnect (connection, FALSE);
+
+  result = dbus_bus_request_name (connection, "org.gnome.GConf",
+                                  0, &bus_error);
+
+  if (dbus_error_is_set (&bus_error))
+    {
+      gconf_log (GCL_WARNING,
+                 _("Failed to get bus name for daemon, exiting: %s"),
+                 bus_error.message);
+      dbus_error_free (&bus_error);
+    }
+
+  if (result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
+    {
+      dbus_connection_unref (connection);
+      return NULL;
+    }
+
+  return connection;
+}
+
+#ifdef ENABLE_DEFAULTS_SERVICE
+/* listen on system bus for defaults changes */
+
+static DBusHandlerResult
+system_bus_message_handler (DBusConnection *connection,
+			    DBusMessage    *message,
+			    void           *user_data)
+{
+  DBusMessage *reply;
+
+  reply = NULL;
+
+  if (dbus_message_is_signal (message,
+			      "org.gnome.GConf.Defaults",
+                              "SystemSet"))
+    {
+      DBusError bus_error;
+      char **keys;
+      int n_keys;
+
+      dbus_error_init (&bus_error);
+      if (dbus_message_get_args (message, &bus_error,
+				 DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &keys, &n_keys,
+				 DBUS_TYPE_INVALID))
+	{
+	  char **key;
+	  GConfSources *system_sources;
+	  GSList addresses;
+
+	  gconf_log (GCL_DEBUG, "System defaults changed.  Notifying.");
+
+	  addresses.data = "xml:merged:/etc/gconf/gconf.xml.system";
+	  addresses.next = NULL;
+	  system_sources = gconf_sources_new_from_addresses (&addresses, NULL);
+
+	  gconfd_clear_cache_for_sources (system_sources);
+
+	  for (key = keys; *key; key++)
+	    gconfd_notify_other_listeners (NULL, system_sources, *key);
+
+	  gconf_sources_free (system_sources);
+
+	  dbus_free_string_array (keys);
+	}
+      else
+        {
+	  gconf_log (GCL_DEBUG, "SystemSet signal received, but error getting message: %s", bus_error.message);
+	}
+      dbus_error_free (&bus_error);
+
+      return DBUS_HANDLER_RESULT_HANDLED;
+    }
+
+  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
+static DBusConnection *
+get_on_system_bus (void)
+{
+  DBusConnection *connection;
+  DBusError bus_error;
+
+  dbus_error_init (&bus_error);
+  connection = dbus_bus_get (DBUS_BUS_SYSTEM, &bus_error);
+
+  if (dbus_error_is_set (&bus_error))
+    {
+      gconf_log (GCL_ERR, _("Could not connect to system bus: %s"), bus_error.message);
+      dbus_error_free (&bus_error);
+      return NULL;
+    }
+
+  dbus_connection_setup_with_g_main (connection, NULL);
+
+  dbus_bus_add_match (connection, "type='signal',interface='org.gnome.GConf.Defaults'", &bus_error);
+  dbus_connection_flush(connection);
+  if (dbus_error_is_set (&bus_error))
+    {
+      gconf_log (GCL_DEBUG, "Failed to add signal match to system bus: %s", bus_error.message);
+      dbus_connection_unref (connection);
+      return NULL;
+    }
+
+  if (!dbus_connection_add_filter (connection, (DBusHandleMessageFunction)
+                                   system_bus_message_handler, NULL, NULL))
+    {
+      gconf_log (GCL_DEBUG, "Failed to add message filter to system bus.");
+      dbus_connection_unref (connection);
+      return NULL;
+    }
+
+  return connection;
+}
+#endif  /* ENABLE_DEFAULTS_SERVICE */
+
+#ifdef G_OS_WIN32
+
 static void
-log_handler (const gchar   *log_domain,
-             GLogLevelFlags log_level,
-             const gchar   *message,
-             gpointer       user_data)
+wait_console_window (void)
 {
-  GConfLogPriority pri = GCL_WARNING;
-  
-  switch (log_level & G_LOG_LEVEL_MASK)
-    {
-    case G_LOG_LEVEL_ERROR:
-    case G_LOG_LEVEL_CRITICAL:
-      pri = GCL_ERR;
-      break;
-
-    case G_LOG_LEVEL_WARNING:
-      pri = GCL_WARNING;
-      break;
-
-    case G_LOG_LEVEL_MESSAGE:
-    case G_LOG_LEVEL_INFO:
-      pri = GCL_INFO;
-      break;
-
-    case G_LOG_LEVEL_DEBUG:
-      pri = GCL_DEBUG;
-      break;
-
-    default:
-      break;
-    }
-
-  gconf_log (pri, "%s", message);
+  SetConsoleTitle ("GConf daemon exiting. Type any character to close this window.");
+  printf ("\n"
+	  "(GConf daemon exiting. Type any character to close this window)\n");
+  _getch ();
 }
 
-/* From ORBit2 */
-/* There is a DOS attack if another user creates
- * the given directory and keeps us from creating
- * it
- */
-static gboolean
-test_safe_tmp_dir (const char *dirname)
-{
-  struct stat statbuf;
-  int fd;
-
-  fd = open (dirname, O_RDONLY);  
-  if (fd < 0)
-    {
-      gconf_log (GCL_WARNING, _("Failed to open %s: %s"),
-                 dirname, g_strerror (errno));
-      return FALSE;
-    }
-  
-  if (fstat (fd, &statbuf) != 0)
-    {
-      gconf_log (GCL_WARNING, _("Failed to stat %s: %s"),
-                 dirname, g_strerror (errno));
-      close (fd);
-      return FALSE;
-    }
-  close (fd);
-  
-  if (statbuf.st_uid != getuid ())
-    {
-      gconf_log (GCL_WARNING, _("Owner of %s is not the current user"),
-                 dirname);
-      return FALSE;
-    }
-  
-  if ((statbuf.st_mode & (S_IRWXG|S_IRWXO)) ||
-      !S_ISDIR (statbuf.st_mode))
-    {
-      gconf_log (GCL_WARNING, _("Bad permissions %lo on directory %s"),
-                 (unsigned long) statbuf.st_mode & 07777, dirname);
-      return FALSE;
-    }
-  
-  return TRUE;
-}
+#endif
 
 int 
 main(int argc, char** argv)
 {
+#ifdef HAVE_SIGACTION
   struct sigaction act;
   sigset_t empty_mask;
+  sigset_t full_mask;
+#endif
   CORBA_Environment ev;
   CORBA_ORB orb;
-  gchar* logname;
-  const gchar* username;
   gchar* ior;
   int exit_code = 0;
   GError *err;
-  char *lock_dir;
-  char *gconfd_dir;
   int dev_null_fd;
   int write_byte_fd;
+  DBusConnection *connection;
 
   _gconf_init_i18n ();
   setlocale (LC_ALL, "");
@@ -593,16 +744,16 @@ main(int argc, char** argv)
   /* This is so we don't prevent unmounting of devices. We divert
    * all messages to syslog
    */
-  if (chdir ("/") < 0)
+  if (g_chdir ("/") < 0)
     {
        g_printerr ("Could not change to root directory: %s\n",
-		g_strerror (errno));
+		   g_strerror (errno));
        exit (1);
     }
 
   if (!g_getenv ("GCONF_DEBUG_OUTPUT"))
     {
-      dev_null_fd = open ("/dev/null", O_RDWR);
+      dev_null_fd = open (DEV_NULL, O_RDWR);
       if (dev_null_fd >= 0)
         {
 	  dup2 (dev_null_fd, 0);
@@ -613,31 +764,52 @@ main(int argc, char** argv)
   else
     {
       gconf_log_debug_messages = TRUE;
+#ifdef G_OS_WIN32
+      if (fileno (stdout) != -1 &&
+	  _get_osfhandle (fileno (stdout)) != -1)
+	{
+	  /* stdout is fine, presumably redirected to a file or pipe */
+	}
+      else
+	{
+	  int allocated_new_console = FALSE;
+
+	  typedef BOOL (* WINAPI AttachConsole_t) (DWORD);
+
+	  AttachConsole_t p_AttachConsole =
+	    (AttachConsole_t) GetProcAddress (GetModuleHandle ("kernel32.dll"), "AttachConsole");
+
+	  if (p_AttachConsole != NULL)
+	    {
+	      if (!AttachConsole (ATTACH_PARENT_PROCESS))
+		{
+		  if (AllocConsole ())
+		    allocated_new_console = TRUE;
+		}
+
+	      freopen ("CONOUT$", "w", stdout);
+	      dup2 (fileno (stdout), 1);
+	      freopen ("CONOUT$", "w", stderr);
+	      dup2 (fileno (stderr), 2);
+
+	      if (allocated_new_console)
+		{
+		  SetConsoleTitle ("GConf daemon debugging output. You can minimize this window, but don't close it.");
+		  printf ("You asked for debugging output by setting the GCONF_DEBUG_OUTPUT\n"
+			  "environment variable, so here it is.\n"
+			  "\n");
+		  atexit (wait_console_window);
+		}
+	    }
+	}
+#endif
     }
   
   umask (022);
   
   gconf_set_daemon_mode(TRUE);
   
-  /* Logs */
-  username = g_get_user_name();
-  logname = g_strdup_printf("gconfd (%s-%u)", username, (guint)getpid());
-
-  openlog (logname, LOG_NDELAY, LOG_USER);
-
-  g_log_set_handler (NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
-                     log_handler, NULL);
-
-  g_log_set_handler ("GLib", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
-                     log_handler, NULL);
-
-  g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
-                     log_handler, NULL);
-  
-  /* openlog() does not copy logname - what total brokenness.
-     So we free it at the end of main() */
-  
-  gconf_log (GCL_INFO, _("starting (version %s), pid %u user '%s'"), 
+  gconf_log (GCL_DEBUG, _("starting (version %s), pid %u user '%s'"), 
              VERSION, (guint)getpid(), g_get_user_name());
 
 #ifdef GCONF_ENABLE_DEBUG
@@ -645,21 +817,29 @@ main(int argc, char** argv)
 #endif
   
   /* Session setup */
+#ifdef HAVE_SIGACTION
+  sigfillset (&full_mask);
+  sigprocmask (SIG_UNBLOCK, &full_mask, NULL);
+
   sigemptyset (&empty_mask);
   act.sa_handler = signal_handler;
   act.sa_mask    = empty_mask;
   act.sa_flags   = 0;
   sigaction (SIGTERM,  &act, NULL);
-  sigaction (SIGILL,  &act, NULL);
-  sigaction (SIGBUS,  &act, NULL);
-  sigaction (SIGFPE,  &act, NULL);
   sigaction (SIGHUP,  &act, NULL);
-  sigaction (SIGSEGV, &act, NULL);
-  sigaction (SIGABRT, &act, NULL);
   sigaction (SIGUSR1,  &act, NULL);
-  
+
   act.sa_handler = SIG_IGN;
   sigaction (SIGINT, &act, NULL);
+#else
+  signal (SIGTERM, signal_handler);
+#ifdef SIGHUP
+  signal (SIGHUP,  signal_handler);
+#endif
+#ifdef SIGUSR1
+  signal (SIGUSR1,  signal_handler);
+#endif
+#endif
 
   CORBA_exception_init(&ev);
 
@@ -686,32 +866,9 @@ main(int argc, char** argv)
   gconf_set_daemon_ior (ior);
   CORBA_free (ior);
 
-  gconfd_dir = gconf_get_daemon_dir ();
-  lock_dir = gconf_get_lock_dir ();
-  
-  if (mkdir (gconfd_dir, 0700) < 0 && errno != EEXIST)
-    gconf_log (GCL_WARNING, _("Failed to create %s: %s"),
-               gconfd_dir, g_strerror (errno));
+  connection = get_on_d_bus ();
 
-  if (!test_safe_tmp_dir (gconfd_dir))
-    {
-      err = g_error_new (GCONF_ERROR,
-                         GCONF_ERROR_LOCK_FAILED,
-                         _("Directory %s has a problem, gconfd can't use it"),
-                         gconfd_dir);
-      daemon_lock = NULL;
-    }
-  else
-    {
-      err = NULL;
-      
-      daemon_lock = gconf_get_lock (lock_dir, &err);
-    }
-
-  g_free (gconfd_dir);
-  g_free (lock_dir);
-
-  if (daemon_lock != NULL)
+  if (connection != NULL)
     {
       /* This loads backends and so on. It needs to be done before
        * we can handle any requests, so before we hit the
@@ -731,18 +888,12 @@ main(int argc, char** argv)
         {
           gconf_log (GCL_ERR, _("Failed to write byte to pipe file descriptor %d so client program may hang: %s"), write_byte_fd, g_strerror (errno));
         }
-
+      
       close (write_byte_fd);
     }
   
-  if (daemon_lock == NULL)
+  if (connection == NULL)
     {
-      g_assert (err);
-
-      gconf_log (GCL_WARNING, _("Failed to get lock for daemon, exiting: %s"),
-                 err->message);
-      g_error_free (err);
-
       enter_shutdown ();
       shutdown_databases ();
       
@@ -751,7 +902,11 @@ main(int argc, char** argv)
 
   /* Read saved log file, if any */
   logfile_read ();
-  
+ 
+#ifdef ENABLE_DEFAULTS_SERVICE 
+  get_on_system_bus ();
+#endif
+
   gconf_main ();
 
   if (in_shutdown)
@@ -785,11 +940,8 @@ main(int argc, char** argv)
 
   daemon_lock = NULL;
   
-  gconf_log (GCL_INFO, _("Exiting"));
+  gconf_log (GCL_DEBUG, _("Exiting"));
 
-  /* Can't do this due to stupid atexit() handler that calls g_log stuff */
-  /*   g_free (logname); */
-  
   return exit_code;
 }
 
@@ -860,12 +1012,12 @@ gconf_main(void)
 
   if (main_loops == NULL)
     {
-      gulong timeout_len = 1000*60*0.5; /* 1 sec * 60 s/min * .5 min */
+      gulong timeout_len = 60*0.5; /* 60 s/min * .5 min */
       
       g_assert(timeout_id == 0);
-      timeout_id = g_timeout_add (timeout_len,
-                                  periodic_cleanup_timeout,
-                                  NULL);
+      timeout_id = g_timeout_add_seconds (timeout_len,
+                                          periodic_cleanup_timeout,
+                                          NULL);
 
     }
   
@@ -1094,8 +1246,15 @@ no_databases_in_use (void)
   /* Only the default database still open, and
    * it has no listeners
    */
-  return db_list == NULL &&
-    gconf_listeners_count (default_db->listeners) == 0;
+
+  if (db_list == NULL)
+    return TRUE;
+
+  if (db_list->next == NULL &&
+      db_list->data == default_db)
+    return gconf_listeners_count (default_db->listeners) == 0;
+
+  return FALSE;
 }
 
 void
@@ -1148,7 +1307,16 @@ gconfd_notify_other_listeners (GConfDatabase *modified_db,
 		      return;
 		    }
 
-		  cvalue = gconf_corba_value_from_gconf_value (value);
+		  if (value != NULL)
+		    {
+		      cvalue = gconf_corba_value_from_gconf_value (value);
+		      gconf_value_free (value);
+		    }
+		  else
+		    {
+		      cvalue = gconf_invalid_corba_value ();
+		    }
+
 		  gconf_database_notify_listeners (db,
 						   NULL,
 						   key,
@@ -1157,7 +1325,6 @@ gconfd_notify_other_listeners (GConfDatabase *modified_db,
 						   is_writable,
 						   FALSE);
 		  CORBA_free (cvalue);
-		  gconf_value_free (value);
 		}
 
 	      tmp2 = tmp2->next;
@@ -1168,6 +1335,22 @@ gconfd_notify_other_listeners (GConfDatabase *modified_db,
     }
 }
 
+void
+gconfd_clear_cache_for_sources (GConfSources *sources)
+{
+  GList *tmp;
+
+  tmp = db_list;
+  while (tmp != NULL)
+    {
+      GConfDatabase *db = tmp->data;
+
+      gconf_database_clear_cache_for_sources (db, sources, NULL);
+
+      tmp = tmp->next;
+    }
+}
+
 /*
  * Cleanup
  */
@@ -1279,7 +1462,7 @@ gconfd_check_in_shutdown (CORBA_Environment *ev)
       ConfigException* ce;
       
       ce = ConfigException__alloc();
-      ce->message = CORBA_string_dup("config server is currently shutting down");
+      ce->message = CORBA_string_dup("Configuration server is currently shutting down");
       ce->err_no = ConfigInShutdown;
 
       CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
@@ -1323,8 +1506,14 @@ gconfd_check_in_shutdown (CORBA_Environment *ev)
 static void
 get_log_names (gchar **logdir, gchar **logfile)
 {
-  *logdir = gconf_concat_dir_and_key (g_get_home_dir (), ".gconfd");
-  *logfile = gconf_concat_dir_and_key (*logdir, "saved_state");
+#ifndef G_OS_WIN32
+      const char *home = g_get_home_dir ();
+#else
+      const char *home = _gconf_win32_get_home_dir ();
+#endif
+
+  *logdir = g_build_filename (home, ".gconfd", NULL);
+  *logfile = g_build_filename (*logdir, "saved_state", NULL);
 }
 
 static void close_append_handle (void);
@@ -1352,18 +1541,18 @@ open_append_handle (GError **err)
 
       get_log_names (&logdir, &logfile);
       
-      mkdir (logdir, 0700); /* ignore failure, we'll catch failures
-                             * that matter on open()
-                             */
+      g_mkdir (logdir, 0700); /* ignore failure, we'll catch failures
+			       * that matter on open()
+			       */
       
-      append_handle = fopen (logfile, "a");
+      append_handle = g_fopen (logfile, "a");
 
       if (append_handle == NULL)
         {
           gconf_set_error (err,
                            GCONF_ERROR_FAILED,
                            _("Failed to open gconfd logfile; won't be able to restore listeners after gconfd shutdown (%s)"),
-                           strerror (errno));
+                           g_strerror (errno));
           
           g_free (logdir);
           g_free (logfile);
@@ -1376,14 +1565,14 @@ open_append_handle (GError **err)
 
 
       {
-        const gulong timeout_len = 1000*60*0.5; /* 1 sec * 60 s/min * 0.5 min */
+        const gulong timeout_len = 60*0.5; /* 60 s/min * 0.5 min */
 
         if (append_handle_timeout != 0)
           g_source_remove (append_handle_timeout);
         
-        append_handle_timeout = g_timeout_add (timeout_len,
-                                               close_append_handle_timeout,
-                                               NULL);
+        append_handle_timeout = g_timeout_add_seconds (timeout_len,
+                                                       close_append_handle_timeout,
+                                                       NULL);
       }
     }
 
@@ -1398,7 +1587,7 @@ close_append_handle (void)
       if (fclose (append_handle) < 0)
         gconf_log (GCL_WARNING,
                    _("Failed to close gconfd logfile; data may not have been properly saved (%s)"),
-                   strerror (errno));
+                   g_strerror (errno));
 
       append_handle = NULL;
 
@@ -1429,9 +1618,9 @@ logfile_save (void)
   
   get_log_names (&logdir, &logfile);
 
-  mkdir (logdir, 0700); /* ignore failure, we'll catch failures
-                         * that matter on open()
-                         */
+  g_mkdir (logdir, 0700); /* ignore failure, we'll catch failures
+			   * that matter on open()
+			   */
 
   saveme = g_string_new (NULL);
 
@@ -1454,13 +1643,13 @@ logfile_save (void)
   /* Now try saving the string to a temporary file */
   tmpfile = g_strconcat (logfile, ".tmp", NULL);
   
-  fd = open (tmpfile, O_WRONLY | O_CREAT | O_TRUNC, 0700);
+  fd = g_open (tmpfile, O_WRONLY | O_CREAT | O_TRUNC, 0700);
 
   if (fd < 0)
     {
       gconf_log (GCL_WARNING,
                  _("Could not open saved state file '%s' for writing: %s"),
-                 tmpfile, strerror (errno));
+                 tmpfile, g_strerror (errno));
       
       goto out;
     }
@@ -1474,49 +1663,56 @@ logfile_save (void)
       
       gconf_log (GCL_WARNING,
                  _("Could not write saved state file '%s' fd: %d: %s"),
-                 tmpfile, fd, strerror (errno));
+                 tmpfile, fd, g_strerror (errno));
 
       goto out;
     }
 
+  if (fsync (fd) < 0)
+    {
+      gconf_log (GCL_WARNING,
+                 _("Could not flush saved state file '%s' to disk: %s"),
+                 tmpfile, g_strerror (errno));
+    }
+
   if (close (fd) < 0)
     {
       gconf_log (GCL_WARNING,
                  _("Failed to close new saved state file '%s': %s"),
-                 tmpfile, strerror (errno));
+                 tmpfile, g_strerror (errno));
       goto out;
     }
 
   fd = -1;
   
   /* Move the main saved state file aside, if it exists */
-  if (gconf_file_exists (logfile))
+  if (g_file_test(logfile, G_FILE_TEST_EXISTS))
     {
       tmpfile2 = g_strconcat (logfile, ".orig", NULL);
-      if (rename (logfile, tmpfile2) < 0)
+      if (g_rename (logfile, tmpfile2) < 0)
         {
           gconf_log (GCL_WARNING,
                      _("Could not move aside old saved state file '%s': %s"),
-                     logfile, strerror (errno));
+                     logfile, g_strerror (errno));
           goto out;
         }
     }
 
   /* Move the new saved state file into place */
-  if (rename (tmpfile, logfile) < 0)
+  if (g_rename (tmpfile, logfile) < 0)
     {
       gconf_log (GCL_WARNING,
-                 _("Failed to move new save state file into place: %s"),
-                 strerror (errno));
+                 _("Failed to move new saved state file into place: %s"),
+                 g_strerror (errno));
 
       /* Try to restore old file */
       if (tmpfile2)
         {
-          if (rename (tmpfile2, logfile) < 0)
+          if (g_rename (tmpfile2, logfile) < 0)
             {
               gconf_log (GCL_WARNING,
                          _("Failed to restore original saved state file that had been moved to '%s': %s"),
-                         tmpfile2, strerror (errno));
+                         tmpfile2, g_strerror (errno));
 
             }
         }
@@ -1526,7 +1722,7 @@ logfile_save (void)
 
   /* Get rid of original saved state file if everything succeeded */
   if (tmpfile2)
-    unlink (tmpfile2);
+    g_unlink (tmpfile2);
   
  out:
   if (saveme)
@@ -1638,7 +1834,7 @@ parse_listener_entry (GHashTable *entries,
   if (err != NULL)
     {
       gconf_log (GCL_DEBUG,
-                 "Failed to unquote config source address from saved state file: %s",
+                 "Failed to unquote configuration source address from saved state file: %s",
                  err->message);
 
       g_error_free (err);
@@ -1857,7 +2053,7 @@ restore_client (const gchar *ior)
   
   if (ev._major != CORBA_NO_EXCEPTION)
     {
-      gconf_log (GCL_DEBUG, "Failed to update client in saved state file, probably the client no longer exists");
+      gconf_log (GCL_DEBUG, "Failed to update client in saved state file, the client probably no longer exists");
 
       goto finished;
     }
@@ -2014,63 +2210,49 @@ restore_client_foreach (gpointer key,
 }
 
 
-#ifndef HAVE_FLOCKFILE
-#  define flockfile(f) (void)1
-#  define funlockfile(f) (void)1
-#  define getc_unlocked(f) getc(f)
-#endif /* !HAVE_FLOCKFILE */
-
 static gchar*
 read_line (FILE *f)
 {
-  int c;
-  GString *str;
-  
-  str = g_string_new (NULL);
-  
-  flockfile (f);
+#define BUF_SIZE 2048
 
-  while (TRUE)
+  char  buf[BUF_SIZE] = { '\0' };
+  char *retval = NULL;
+  int   len = 0;
+
+  do
     {
-      c = getc_unlocked (f);
-
-      switch (c)
+      if (fgets (buf, BUF_SIZE, f) == NULL)
         {
-        case EOF:
           if (ferror (f))
             {
               gconf_log (GCL_ERR,
                          _("Error reading saved state file: %s"),
                          g_strerror (errno));
             }
-
-          /* FALL THRU */
-        case '\n':
-          goto done;
-          break;
-
-        default:
-          g_string_append_c (str, c);
           break;
         }
-    }
 
- done:
-  funlockfile (f);
+      len = strlen (buf);
+      if (buf[len - 1] == '\n')
+	buf[--len] = '\0';
 
-  if (str->len == 0)
-    {
-      g_string_free (str, TRUE);
-      return NULL;
-    }
-  else
-    {
-      gchar *ret;
-      
-      ret = str->str;
-      g_string_free (str, FALSE);
-      return ret;
+      if (retval == NULL)
+	{
+	  retval = g_strndup (buf, len);
+	}
+      else
+	{
+	  char *freeme = retval;
+
+	  retval = g_strconcat (retval, buf, NULL);
+	  g_free (freeme);
+	}
     }
+  while (len == BUF_SIZE - 1);
+
+  return retval;
+
+#undef BUF_SIZE
 }
 
 static void
@@ -2089,7 +2271,7 @@ logfile_read (void)
   
   get_log_names (&logdir, &logfile);
 
-  f = fopen (logfile, "r");
+  f = g_fopen (logfile, "r");
   
   if (f == NULL)
     {
@@ -2270,7 +2452,7 @@ log_client_change (const ConfigListener client,
     {
       gconf_log (GCL_WARNING,
                  _("Failed to write client add to saved state file: %s"),
-                 strerror (errno));
+                 g_strerror (errno));
       goto error;
     }
 
@@ -2278,7 +2460,7 @@ log_client_change (const ConfigListener client,
     {
       gconf_log (GCL_WARNING,
                  _("Failed to flush client add to saved state file: %s"),
-                 strerror (errno));
+                 g_strerror (errno));
       goto error;
     }
 
@@ -2498,10 +2680,3 @@ client_count (void)
     return g_hash_table_size (client_table);
 }
 
-
-static void
-gconf_handle_segv (int signum)
-{
-  return;
-}
-
